/*
 * HFS (aka MAC) filesystem/handler for MorphOS/AmigaOS
 * Author: Marek 'March' Szyprowski <marek@amiga.pl>
 */

#include <exec/types.h>
#include <dos/dos.h>
#include <devices/trackdisk.h>
#include <devices/inputevent.h>

#include <proto/exec.h>
#include <proto/dos.h>

#include "mac_fs.h"
#include "mac_protos.h"

LONG FreeUnusedDisk(struct FSSuper *sb, struct FSSuper *prev)
{
	if (sb->doslist->dol_misc.dol_volume.dol_LockList == NULL) /* check if the device can be removed */
	{
		kprintf("\tDisk is no longer in use - removing it completely\n");
		SendDiskChangePacket(ACTION_VOLUME_REMOVE, sb->doslist);
		sb->doslist = NULL;
		FreeSuper(sb);
		if (prev)
			prev->next = sb->next;
		else
			glob->sblist = sb->next;
		FS_FreeMem(sb);
		return 0;
	}
	return ERROR_OBJECT_IN_USE;
}

void SendDiskChangePacket(ULONG action, APTR data)
{
	struct DosPacket *dospacket;

	dospacket = AllocDosObject(DOS_STDPKT, TAG_DONE);
	dospacket->dp_Type = ACTION_DISK_CHANGE;
	dospacket->dp_Arg1 = ID_HFS_DISK;
	dospacket->dp_Arg2 = (ULONG) data;
	dospacket->dp_Arg3 = action;
	dospacket->dp_Port = NULL;

	PutMsg(glob->ourport, dospacket->dp_Link);
}

void DoDiskInsert()
{
	struct FSSuper *sb;

	if (glob->sb == NULL && (sb = FS_AllocMem(sizeof(struct FSSuper))))
	{
		memset(sb, 0, sizeof(struct FSSuper));
		if (glob->autodetect)
			UpdateDiskGeometry();

		if (ReadSuper(sb) == 0)
		{
			LONG found = FALSE;
			if (glob->sblist)
			{
				struct FSSuper *ptr = glob->sblist, *prev = NULL;
				while (ptr != NULL)
				{
					if (CompareSuper(sb, ptr) == 0)
						break;
					prev = ptr;
					ptr = ptr->next;
				}
				if (ptr)
				{
					kprintf("\tFound Super Block in spare list, freeing obsolete old one\n");

					sb->doslist = ptr->doslist;
					ptr->doslist = NULL;
					sb->doslist->dol_Name = (BSTR)MKBADDR(&sb->name);

					if (prev)
						prev->next = ptr->next;
					else
						glob->sblist = ptr->next;

					FreeSuper(ptr);
					FS_FreeMem(ptr);

					RestoreLockStructures(sb);
					found = TRUE;
				}
			}
			if (!found)
			{
				struct DosList *newvol;
				kprintf("\tCreating new volume.\n");
				if ((newvol = FS_AllocMem(sizeof(struct DosList))))
				{
					newvol->dol_Next = NULL;
					newvol->dol_Type = DLT_VOLUME;
					newvol->dol_Task = glob->ourport;
					newvol->dol_Lock = NULL;

					MacTime2DateStamp(sb->create_date, &newvol->dol_misc.dol_volume.dol_VolumeDate);
					newvol->dol_misc.dol_volume.dol_LockList = NULL;
					newvol->dol_misc.dol_volume.dol_DiskType = ID_HFS_DISK;

					newvol->dol_Name = (BSTR)MKBADDR(&sb->name);
					sb->doslist = newvol;

					SendDiskChangePacket(ACTION_VOLUME_ADD, newvol);
				}
			}
			else
				SendEvent(IECLASS_DISKINSERTED);
			glob->sb = sb;
			kprintf("\tDisk successfully initialised\n");
			return;
		}
		FS_FreeMem(sb);
	}
	SendEvent(IECLASS_DISKINSERTED);
	return;
}

void DoDiskRemove()
{
	if (glob->sb)
	{
		struct FSSuper *sb = glob->sb;
		glob->sb = NULL;
		if (sb->doslist->dol_misc.dol_volume.dol_LockList == NULL) /* check if the device can be removed */
		{
			kprintf("\tRemoving disk completely\n");
			SendDiskChangePacket(ACTION_VOLUME_REMOVE, sb->doslist);
			sb->doslist = NULL;
			FreeSuper(sb);
			FS_FreeMem(sb);
		}
		else
		{
			sb->next = glob->sblist;
			glob->sblist = sb;
			FreeLockStructures(sb);
			kprintf("\tMoved in-memory super block to spare list. Waiting for locks to be freed\n");
			SendEvent(IECLASS_DISKREMOVED);
		}
	}
}
 
void ProcessDiskChange ()
{
	kprintf("\n");
	kprintf("Got disk change request\n");
	
	if (glob->disk_inhibited)
	{
		kprintf("Disk is inhibited, ignoring disk change\n");
		return;
	}

	glob->diskioreq->iotd_Req.io_Command = TD_CHANGESTATE;
	glob->diskioreq->iotd_Req.io_Data = NULL;
	glob->diskioreq->iotd_Req.io_Length = 0;
	glob->diskioreq->iotd_Req.io_Flags = IOF_QUICK;
	DoIO((struct IORequest*)glob->diskioreq);

	if (glob->diskioreq->iotd_Req.io_Error == 0 && glob->diskioreq->iotd_Req.io_Actual == 0)
	{
		/* Disk has been inserted. */
		kprintf("\tDisk has been inserted\n");
		glob->disk_inserted = TRUE;
		DoDiskInsert();
	}
	else
	{
		/* Disk has been removed. */
		kprintf("\tDisk has been removed\n");
		glob->disk_inserted = FALSE;
		DoDiskRemove();
	}
	kprintf("Done\n");
}

